{{
        VGA 640 x 480 2-bit with 4-bit inset driver.
           (c) Copyright 2012 Philip C. Pilgrim
         "vga_driver" portion loosely derived from
           VGA64_PIXEngine by Kwabena W. Agyeman
}}

CON

  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000

'Size of the color palette FIFO buffer.

  BUF_SIZE      = 511

'Pin groups.
  
  PROP_BOE_0    = 0 'PO - P7 
  PROP_BOE_1    = 1 'P8 - P15
  DEMO_BOARD    = 2 'P16 - P23

VAR

  long  color_buf[BUF_SIZE], deq_ptr
  byte  cog_no

PUB main

  start(DEMO_BOARD)

PUB start(pin_group) | pix_clk

  pin_group := ((pin_group <# 3) #> 0)
  dira1 := dira0 := ($FF << (pin_group << 3))
  vcfg0 := %0_01_0 << 28 | pin_group << 9 | $ff
  vcfg1 := %0_01_1 << 28 | pin_group << 9 | $ff
  cvsync_mask := vsync_mask := 1 << (pin_group << 3)
  hsync_mask := vsync_mask << 1
   
  pix_clk := constant(25_175_000 / 4)
  frqa0 := 1
   
  repeat 32
    pix_clk <<= 1
    frqa0 <-= 1
    if(pix_clk => clkfreq)
      pix_clk -= clkfreq
      frqa0++
   
  frqa1 := frqa0
  buffer_addr := @color_buf
  deq_ptr := @color_buf
  longfill(@color_buf, $fc_00_54_a8, BUF_SIZE)
  cognew(@vga_driver, @reg_data)
  cognew(@ovl_cdriver, @buffer_addr)
  cognew(@ovl_pdriver, @buffer_addr)

DAT

'-------[ 640 x 480 x 1-bit VGA Driver ]---------------------------------------

              org       0

vga_driver    mov       vcfg,vcfg0              ' Setup video hardware.
              mov       frqa,frqa0         
              movi      ctra,#%0_00001_101
              mov       dira,dira0      

'       Active Video

frame_loop    mov       line_ctr,#480
              mov       disp_ptr,par
              rdlong    y_alt,disp_ptr
              add       disp_ptr,#4
              mov       y_ctr,#1

line_loop     djnz      y_ctr,#intra_block

              mov       y_ctr,#15
              mov       alt_master,#0
              shl       y_alt,#1 wc
        if_nc jmp       #intra_block
        
              rdlong    alt_master,disp_ptr
              add       disp_ptr,#4
              mov       alt_colors,alt_master
              shr       alt_colors,#4
              or        alt_colors,#3
              shl       alt_colors,#8
              mov       pixels,alt_master
              and       pixels,#%111111
              shl       pixels,#2
              or        alt_colors,pixels
              or        alt_colors,#3
        
intra_block   mov       vscl,vis_vscl           ' Set the video scale.
              rdlong    reg_mask,disp_ptr
              add       disp_ptr,#4
              mov       reg_colors,reg_mask
              shr       reg_colors,#4
              or        reg_colors,#3
              shl       reg_colors,#8
              mov       pixels,reg_mask
              and       pixels,#%111111
              shl       pixels,#2
              or        reg_colors,pixels
              or        reg_colors,#3
              mov       alt_mask,alt_master
              mov       counter,#20             ' 20 x 32-bit tiles = 640 pixels

'       Visible Video

pix_loop      shl       reg_mask,#1 wc
        if_c  rdlong    pixels,disp_ptr
        if_c  add       disp_ptr,#4
        if_nc mov       pixels,#0
              shl       alt_mask,#1 wc
        if_c  waitvid   alt_colors,pixels
        if_nc waitvid   reg_colors,pixels
              djnz      counter,#pix_loop       ' Repeat.

'       Horizontal Front Porch, Sync, and Back Porch

              mov       vscl,invis_vscl         ' Set/Reset the video scale.
              waitvid   hsync_colors,sync_pix   ' Horizontal Sync.
              djnz      line_ctr,#line_loop 

'       Vertical Front Porch

              mov       counter,#10             ' Set loop counter.

front_porch   mov       vscl,blank_vscl         ' Invisible lines.
              waitvid   hsync_colors,#0                         '

              mov       vscl,invis_vscl         ' Horizontal Sync.
              waitvid   hsync_colors,sync_pix    
              djnz      counter,#front_porch    ' Repeat # times.

'       Vertical Sync

              mov       counter,#4              ' Set loop counter.

vsync_loop    mov       vscl,blank_vscl         ' Invisible lines.
              waitvid   vsync_colors,#0          

              mov       vscl,invis_vscl         ' Vertical sync.
              waitvid   vsync_colors,sync_pix    
              djnz      counter,#vsync_loop     ' Repeat.

'       Vertical Back Porch

              mov       counter,#31             ' Set loop counter.

back_porch    mov       vscl,blank_vscl         ' Invisible lines.
              waitvid   hsync_colors,#0          

              mov       vscl,invis_vscl         ' Horizontal Sync.
              waitvid   hsync_colors,sync_pix    
              djnz      counter,#back_porch     ' Repeat # times.

              jmp       #frame_loop             ' Loop.

'       Data

invis_vscl    long      (16 << 12) + 160        ' Scaling for inactive video.
vis_vscl      long      (1 << 12) + 32          ' Scaling for active video.
blank_vscl    long      640                     ' Blank scanline pixel length.
sync_pix      long      $00_00_00_7e            ' F-porch, h-sync, and b-porch.
hsync_colors  long      $01_03_01_03            ' Horizontal sync color mask.
vsync_colors  long      $00_02_00_02            ' Vertical sync color mask.

'       Configuration Settings

dira0         long      0-0
vcfg0         long      0-0
frqa0         long      0-0

'       Run Time Variables

counter       res       1
pixels        res       1
tile_ctr      res       1
line_ctr      res       1
disp_ptr      res       1
alt_master    res       1
reg_mask      res       1
alt_mask      res       1
reg_colors    res       1
alt_colors    res       1
y_alt         res       1
y_ctr         res       1

'-------[ Two-bit VGA Overlay Driver ]-----------------------------------------

              org       0

ovl_pdriver   rdlong    buf_addr,par
              mov       deq_addr,#BUF_SIZE
              shl       deq_addr,#2
              add       deq_addr,buf_addr
              mov       buf_index,buf_addr

              mov       pix_addr,par
              add       pix_addr,#8
             
              mov       frqa,frqa1         
              movi      ctra,#%0_00001_101
              mov       dira,dira1
              mov       vscl,ovl_vscl
              
vsync_wait    waitpne   vsync_mask,vsync_mask
              mov       buf_index,buf_addr
              wrlong    buf_index,deq_addr
              waitpeq   vsync_mask,vsync_mask
              mov       hdelay,top
              shl       hdelay,#4
              sub       hdelay,top
              add       hdelay,#30
              mov       pix_ptr,pix_addr

hsync_wait    waitpne   hsync_mask,hsync_mask
              waitpeq   hsync_mask,hsync_mask
              djnz      hdelay,#hsync_wait

              mov       hdelay,height
              shl       hdelay,#4
              sub       hdelay,height
              
line_lp       waitpne   hsync_mask,hsync_mask
              mov       phsa,#0
              mov       pix_cnt,cnt
              add       pix_cnt,hwait
              waitcnt   pix_cnt,#0
              mov       vcfg,vcfg1
              mov       pix_cnt,left
           
blank_lp      waitvid   zero,#0
              djnz      pix_cnt,#blank_lp

              mov       pix_cnt,width

pix_lp        rdlong    ovl_pixels,pix_ptr
              add       pix_ptr,#4
              rdlong    ovl_colors,buf_index
              add       buf_index,#4
              cmp       buf_index,deq_addr wz
        if_z  mov       buf_index,buf_addr
              wrlong    buf_index,deq_addr
              waitvid   ovl_colors,ovl_pixels
              djnz      pix_cnt,#pix_lp

              waitvid   zero,#0

              mov       vcfg,#0
              djnz      hdelay,#line_lp
              
              jmp       #vsync_wait

dira1         long      0-0
vcfg1         long      0-0
frqa1         long      0-0
hsync_mask    long      0-0
vsync_mask    long      0-0

ovl_colors    long      $fc_00_54_a8 '$fc_a8_54_00
ovl_pixels    long      0 '%%0000_1111_2222_3333
nul_colors    long      0
ovl_vscl      long      (2 << 12) + 32
hwait         long      343
zero          long      0
buf_index     long      0

top           long      6
left          long      1
width         long      9
height        long      22

hdelay        res       1
pix_cnt       res       1
pix_addr      res       1
pix_ptr       res       1
buf_addr      res       1
deq_addr      res       1

'-------[ Overlay color unpacker ]---------------------------------------------

              org       0

ovl_cdriver   rdlong    cbuf_addr,par           'Get the start address of the FIFO color buffer.
              mov       cdeq_addr,#BUF_SIZE     'Compute the address of the dequeue pointer.
              shl       cdeq_addr,#2
              add       cdeq_addr,cbuf_addr
              mov       cbuf_index,cbuf_addr

              mov       cfile_addr,par          'Get the starting address of the overlay file.
              add       cfile_addr,#4           
              rdlong    cacc,cfile_addr         'Read the offset for the color swap data.
              add       cfile_addr,cacc         'Add offset to address.

cmain_lp      waitpeq   cvsync_mask,cvsync_mask
              mov       cfile_ptr,cfile_addr    'Move pointer to the beginning of the color swap data.
              mov       palette,#0              'Initialize palette to black.
              neg       color_ptr,#1            'Initialize color point to ficticious last color.
              mov       cbuf_index,cbuf_addr
              rdlong    test_pattern,cfile_ptr

color_lp      rdbyte    color,cfile_ptr         'Read next color swap byte.
              add       cfile_ptr,#1            'Increment pointer.
              shr       color,#1 wc             'Add 2 LSBs of color + 1 to color pointer
              addx      color_ptr,#1
              shr       color,#1 wc
        if_c  add       color_ptr,#2
              cmpsub    color_ptr,#4 wc         'Is it beyond this palette?
        if_nc jmp       #continue               '  No:  Process it.

              mov       cacc,cbuf_index         '  Yes: Calculate next buffer index.
              add       cacc,#4
              cmp       cacc,cdeq_addr wz       '       Over the top?
        if_z  mov       cacc,cbuf_addr          '         Yes: Reset to beginning.
        
cbuf_wait     rdlong    caccx,cdeq_addr          '       Get the dequeue pointer.
              cmp       caccx,cacc wz           '       Equal to new buffer index?
        if_z  jmp       #cbuf_wait              '         Yes: Buffer is full. Try again.

              wrlong    palette,cbuf_index      '       Write palette to buffer.
              mov       cbuf_index,cacc         '       Update buffer index with new value.

continue      mov       cacc,color_ptr          'Compute shift amount from pointer.
              shl       cacc,#3                 ' x 8
              add       cacc,#2                 ' + 2
              shl       color,cacc              'Shift color into position.
              xor       palette,color           'XOR it into palette.
              
              test      cvsync_mask,ina wz
        if_nz jmp       #color_lp
        
              jmp       #cmain_lp               

cbuf_index    long      0
cvsync_mask   long      0-0

test_pattern  long      $fc_c0_30_0c

cbuf_addr     res       1
cdeq_addr     res       1
cfile_addr    res       1
cfile_ptr     res       1
cacc          res       1
caccx         res       1
palette       res       1
color_ptr     res       1
color         res       1


'-------[ File data and hub address info ]-------------------------------------

reg_data      file      "slide17_text.dat"
buffer_addr   long      0-0
ovl_data      file      "boebot.dat"

''-----------[ TERMS OF USE ]---------------------------------------------------
''
'' Permission is hereby granted, free of charge, to any person obtaining a copy of
'' this software and associated documentation files (the "Software"), to deal in
'' the Software without restriction, including without limitation the rights to use,
'' copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
'' Software, and to permit persons to whom the Software is furnished to do so,
'' subject to the following conditions: 
''
'' The above copyright notice and this permission notice shall be included in all
'' copies or substantial portions of the Software. 
''
'' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
'' INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
'' PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
'' HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
'' OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'' SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

